#include "MemoryNameSystem.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/ManagedIceObject.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/Writer.h>

#include <SimoxUtility/algorithm/string/string_tools.h>

namespace armarx::armem::client
{

    MemoryNameSystem::MemoryNameSystem()
    {
    }


    MemoryNameSystem::MemoryNameSystem(mns::MemoryNameSystemInterfacePrx mns, ManagedIceObject* component) :
        mns(mns), component(component)
    {
    }


    void MemoryNameSystem::update()
    {
        ARMARX_CHECK_NOT_NULL(mns);

        data::GetAllRegisteredMemoriesResult result;
        try
        {
            result = mns->getAllRegisteredMemories();
        }
        catch (const Ice::NotRegisteredException& e)
        {
            throw error::MemoryNameSystemQueryFailed(e.what());
        }

        if (result.success)
        {
            for (const auto& [name, proxy] : result.proxies)
            {
                auto [it, inserted] = servers.try_emplace(name, proxy);
                // inserted: OK
                if (not inserted)
                {
                    // Compare ice identities, update if it changed.
                    if (it->second->ice_getIdentity() != proxy->ice_getIdentity())
                    {
                        // Replace old proxy by new one.
                        it->second = proxy;
                    }
                }
            }
            // Remove all entries which are not there anymore.
            for (auto it = servers.begin(); it != servers.end();)
            {
                if (result.proxies.count(it->first) == 0)
                {
                    it = servers.erase(it);
                }
                else
                {
                    ++it;
                }
            }
        }
        else
        {
            throw error::MemoryNameSystemQueryFailed(result.errorMessage);
        }
    }


    server::MemoryInterfacePrx MemoryNameSystem::resolveServer(const MemoryID& memoryID)
    {
        if (auto it = servers.find(memoryID.memoryName); it != servers.end())
        {
            return it->second;
        }
        else
        {
            update();
            if (auto it = servers.find(memoryID.memoryName); it != servers.end())
            {
                return it->second;
            }
            else
            {
                throw error::CouldNotResolveMemoryServer(memoryID);
            }
        }
    }


    server::MemoryInterfacePrx MemoryNameSystem::waitForServer(const MemoryID& memoryID, Time timeout)
    {
        if (auto it = servers.find(memoryID.memoryName); it != servers.end())
        {
            return it->second;
        }
        else
        {
            armem::data::WaitForMemoryInput input;
            input.name = memoryID.memoryName;
            input.timeoutMilliSeconds = timeout.toMilliSeconds();

            ARMARX_CHECK_NOT_NULL(mns);
            armem::data::WaitForMemoryResult result = mns->waitForMemory(input);
            if (result.success)
            {
                return result.proxy;
            }
            else
            {
                throw error::CouldNotResolveMemoryServer(memoryID, result.errorMessage);
            }
        }
    }

    server::MemoryInterfacePrx MemoryNameSystem::useServer(const MemoryID& memoryID)
    {
        ARMARX_CHECK_NOT_NULL(component)
                << "Owning component not set when using a memory server. \n"
                << "When calling `armem::mns::MemoryNameSystem::useServer()`, the owning component which should "
                << "receive the dependency to the memory server must be set beforehand. \n\n"
                << "Use `armem::mns::MemoryNameSystem::setComponent()` or pass the component on construction "
                << "before calling useServer().";
        return useServer(memoryID, *component);
    }


    server::MemoryInterfacePrx MemoryNameSystem::useServer(const MemoryID& memoryID, ManagedIceObject& component)
    {
        server::MemoryInterfacePrx server = waitForServer(memoryID);
        // Add dependency.
        component.usingProxy(server->ice_getIdentity().name);
        return server;
    }


    server::MemoryInterfacePrx MemoryNameSystem::useServer(const std::string& memoryName)
    {
        return useServer(MemoryID().withMemoryName(memoryName));
    }


    server::MemoryInterfacePrx MemoryNameSystem::useServer(const std::string& memoryName, ManagedIceObject& component)
    {
        return useServer(MemoryID().withMemoryName(memoryName), component);
    }


    Reader MemoryNameSystem::getReader(const MemoryID& memoryID)
    {
        return Reader(resolveServer(memoryID));
    }


    Reader MemoryNameSystem::useReader(const MemoryID& memoryID)
    {
        return Reader(useServer(memoryID));
    }


    Reader MemoryNameSystem::useReader(const MemoryID& memoryID, ManagedIceObject& component)
    {
        return Reader(useServer(memoryID, component));
    }


    Reader MemoryNameSystem::useReader(const std::string& memoryName)
    {
        return useReader(MemoryID().withMemoryName(memoryName));
    }


    Reader MemoryNameSystem::useReader(const std::string& memoryName, ManagedIceObject& component)
    {
        return useReader(MemoryID().withMemoryName(memoryName), component);
    }


    template <class ClientT>
    std::map<std::string, ClientT> MemoryNameSystem::_getAllClients() const
    {
        std::map<std::string, ClientT> result;
        for (const auto& [name, server] : servers)
        {
            result[name] = ClientT(server);
        }
        return result;
    }


    template <class ClientT>
    std::map<std::string, ClientT> MemoryNameSystem::_getAllClients(bool update)
    {
        if (update)
        {
            this->update();
        }
        return const_cast<const MemoryNameSystem&>(*this)._getAllClients<ClientT>();
    }


    std::map<std::string, Reader> MemoryNameSystem::getAllReaders(bool update)
    {
        return _getAllClients<Reader>(update);
    }


    std::map<std::string, Reader> MemoryNameSystem::getAllReaders() const
    {
        return _getAllClients<Reader>();
    }


    Writer MemoryNameSystem::getWriter(const MemoryID& memoryID)
    {
        return Writer(resolveServer(memoryID));
    }


    Writer MemoryNameSystem::useWriter(const MemoryID& memoryID)
    {
        return Writer(useServer(memoryID));
    }


    Writer MemoryNameSystem::useWriter(const MemoryID& memoryID, ManagedIceObject& component)
    {
        return Writer(useServer(memoryID, component));
    }


    Writer MemoryNameSystem::useWriter(const std::string& memoryName)
    {
        return useWriter(MemoryID().withMemoryName(memoryName));
    }


    Writer MemoryNameSystem::useWriter(const std::string& memoryName, ManagedIceObject& component)
    {
        return useWriter(MemoryID().withMemoryName(memoryName), component);
    }


    std::map<std::string, Writer> MemoryNameSystem::getAllWriters(bool update)
    {
        return _getAllClients<Writer>(update);
    }


    std::map<std::string, Writer> MemoryNameSystem::getAllWriters() const
    {
        return _getAllClients<Writer>();
    }


    std::optional<wm::EntityInstance> MemoryNameSystem::resolveEntityInstance(const MemoryID& id)
    {
        auto result = resolveEntityInstances({id});
        if (result.size() > 0)
        {
            return result.begin()->second;
        }
        else
        {
            return std::nullopt;
        }
    }


    std::map<MemoryID, wm::EntityInstance> MemoryNameSystem::resolveEntityInstances(const std::vector<MemoryID>& ids)
    {
        std::stringstream errors;
        int errorCounter = 0;

        std::map<std::string, std::vector<MemoryID>> idsPerMemory;
        for (const auto& id : ids)
        {
            idsPerMemory[id.memoryName].push_back(id);
        }

        std::map<MemoryID, wm::EntityInstance> result;
        for (const auto& [memoryName, ids] : idsPerMemory)
        {
            Reader reader = getReader(MemoryID().withMemoryName(memoryName));
            QueryResult queryResult = reader.queryMemoryIDs(ids);
            if (queryResult.success)
            {
                for (const MemoryID& id : ids)
                {
                    try
                    {
                        if (id.hasInstanceIndex())
                        {
                            result[id] = queryResult.memory.getInstance(id);
                        }
                        else if (id.hasTimestamp())
                        {
                            result[id] = queryResult.memory.getSnapshot(id).getInstance(0);
                        }
                        else if (id.hasEntityName())
                        {
                            result[id] = queryResult.memory.getEntity(id).getLatestSnapshot().getInstance(0);
                        }
                        else
                        {
                            std::stringstream ss;
                            ss << "MemoryNameSystem::" << __FUNCTION__ << "requires IDs to be entity, snapshot or instance IDs,"
                               << "but ID has no entity name.";
                            throw error::InvalidMemoryID(id, ss.str());
                        }
                    }
                    catch (const error::ArMemError& e)
                    {
                        errors << "\n#" << ++errorCounter << "\n"
                               << "Failed to retrieve " << id << " from query result: \n" << e.what();
                    }
                }
            }
            else
            {
                errors << "\n# " << ++errorCounter << "\n"
                       << "Failed to query '" << memoryName << "': \n" << queryResult.errorMessage;
            }
        }

        if (errors.str().size() > 0)
        {
            ARMARX_INFO << "MemoryNameSystem::" << __FUNCTION__ << ": The following errors may affect your result: "
                        << "\n\n" << errors.str() << "\n\n"
                        << "When querying entity instances: \n- "
                        << simox::alg::join(simox::alg::multi_to_string(ids), "\n- ");
        }

        return result;
    }


    void MemoryNameSystem::registerServer(const MemoryID& memoryID, server::MemoryInterfacePrx proxy)
    {
        data::RegisterMemoryInput input;
        input.name = memoryID.memoryName;
        input.proxy = proxy;
        ARMARX_CHECK_NOT_NULL(input.proxy);

        ARMARX_CHECK_NOT_NULL(mns);
        data::RegisterMemoryResult result = mns->registerMemory(input);
        if (!result.success)
        {
            throw error::ServerRegistrationOrRemovalFailed("register", memoryID, result.errorMessage);
        }
    }


    void MemoryNameSystem::removeServer(const MemoryID& memoryID)
    {
        data::RemoveMemoryInput input;
        input.name = memoryID.memoryName;

        ARMARX_CHECK_NOT_NULL(mns);
        data::RemoveMemoryResult result = mns->removeMemory(input);
        if (!result.success)
        {
            throw error::ServerRegistrationOrRemovalFailed("remove", memoryID, result.errorMessage);
        }
    }


    mns::MemoryNameSystemInterfacePrx MemoryNameSystem::getMemoryNameSystem() const
    {
        return mns;
    }


    void MemoryNameSystem::getMemoryNameSystem(mns::MemoryNameSystemInterfacePrx mns)
    {
        this->mns = mns;
    }

    void MemoryNameSystem::setComponent(ManagedIceObject* component)
    {
        this->component = component;
    }

}



