#include "TypedDataDisplayVisitor.h"

#include <iomanip>      // std::setprecision

#include <SimoxUtility/algorithm/string.h>
#include <SimoxUtility/math/pose/pose.h>

#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/converter/eigen/EigenConverter.h>
#include <RobotAPI/libraries/armem/core/Time.h>

#include "DataDisplayVisitor.h"


namespace armarx::aron
{

    std::string TypedDataDisplayVisitor::getValue(TypeNavigator& type, DataNavigator& data)
    {
        TypedDataDisplayVisitor v;
        bool r = v.applyTo(type, data);
        ARMARX_CHECK(!r);
        return v.value.str();
    }


    TypedDataDisplayVisitor::TypedDataDisplayVisitor() :
        coeffSep("  "),
        eigenIof(Eigen::StreamPrecision, 0, coeffSep, "\n", "", "", "", "")
    {
    }


    bool TypedDataDisplayVisitor::visitEnter(DictTypeNavigator&, DictDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visitEnter(ObjectTypeNavigator&, DictDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visitEnter(ListTypeNavigator&, ListDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visitEnter(TupleTypeNavigator&, ListDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(BoolTypeNavigator&, BoolDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(DoubleTypeNavigator&, DoubleDataNavigator& data)
    {
        setStreamPrecision();
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(FloatTypeNavigator&, FloatDataNavigator& data)
    {
        setStreamPrecision();
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(IntTypeNavigator&, IntDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(LongTypeNavigator&, LongDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(StringTypeNavigator&, StringDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(TimeTypeNavigator&, LongDataNavigator& data)
    {
        armem::Time time = armem::Time::microSeconds(data.getValue());
        value << armem::toDateTimeMilliSeconds(time);
        return false;
    }


    template <typename ScalarT>
    void TypedDataDisplayVisitor::processEigenMatrix(EigenMatrixTypeNavigator& type, NDArrayDataNavigator& data)
    {
        Eigen::Map<Eigen::Matrix<ScalarT, Eigen::Dynamic, Eigen::Dynamic>> m(reinterpret_cast<ScalarT*>(data.getData()), type.getRows(), type.getCols());
        value << m.format(eigenIof);
    }

    bool TypedDataDisplayVisitor::visit(EigenMatrixTypeNavigator& t, NDArrayDataNavigator& data)
    {
        if (std::max(t.getRows(), t.getCols()) > 10)
        {
            // Just show the shape.
            value << DataDisplayVisitor::getValue(data);
        }
        else if (data.getType() == "float")
        {
            setStreamPrecision();
            processEigenMatrix<float>(t, data);
        }
        else if (data.getType() == "double")
        {
            setStreamPrecision();
            processEigenMatrix<double>(t, data);
        }
        else
        {
            value << DataDisplayVisitor::getValue(data);
        }
        return false;
    }

    bool TypedDataDisplayVisitor::visit(EigenQuaternionTypeNavigator&, NDArrayDataNavigator& data)
    {
        processEigenQuaternion(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(IVTCByteImageTypeNavigator&, NDArrayDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(OpenCVMatTypeNavigator&, NDArrayDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(PCLPointCloudTypeNavigator&, NDArrayDataNavigator& data)
    {
        value << DataDisplayVisitor::getValue(data);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(PoseTypeNavigator&, NDArrayDataNavigator& data)
    {
        const Eigen::IOFormat eigenIof(Eigen::StreamPrecision, 0, " ", "\n", "( ", " )", "", "");
        const std::string cdot = "\u00B7";  // center dot: https://unicode-table.com/de/00B7/

        auto getLines = [&](auto&& mat)
        {
            std::stringstream ss;
            setStreamPrecision(ss);
            ss << mat.format(eigenIof);
            std::vector<std::string> lines = simox::alg::split(ss.str(), "\n");
            ARMARX_CHECK_EQUAL(lines.size(), 3);
            return lines;
        };

        const Eigen::Matrix4f pose = aron::converter::AronEigenConverter::ConvertToMatrix4f(data);
        const std::vector<std::string> r = getLines(simox::math::orientation(pose));
        const std::vector<std::string> t = getLines(simox::math::position(pose));

        std::vector<std::string> lines;
        lines.push_back(r.at(0) + "      \t" + t.at(0));
        lines.push_back(r.at(1) + " " + cdot + " x + \t" + t.at(1));
        lines.push_back(r.at(2) + "      \t" + t.at(2));

        value << simox::alg::join(lines, "\n");
        return false;
    }


    bool TypedDataDisplayVisitor::visit(PositionTypeNavigator&, NDArrayDataNavigator& data)
    {
        const Eigen::Vector3f pos = aron::converter::AronEigenConverter::ConvertToVector3f(data);
        setStreamPrecision();
        value << pos.format(eigenIof);
        return false;
    }

    bool TypedDataDisplayVisitor::visit(OrientationTypeNavigator&, NDArrayDataNavigator& data)
    {
        processEigenQuaternion(data);
        return false;
    }

    void TypedDataDisplayVisitor::processEigenQuaternion(visitor::TypedDataVisitor::NDArrayDataNavigator& data)
    {
        const Eigen::Quaternionf quat = aron::converter::AronEigenConverter::ConvertToQuaternionf(data);
        setStreamPrecision();
        value << quat.w() << coeffSep << "|" << coeffSep << quat.x() << coeffSep << quat.y() << coeffSep << quat.z();
    }

    void TypedDataDisplayVisitor::setStreamPrecision()
    {
        setStreamPrecision(value);
    }

    void TypedDataDisplayVisitor::setStreamPrecision(std::ostream& os)
    {
        os << std::setprecision(2) << std::fixed;
    }

}
