/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::armem_gui
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::armem_gui

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
// #include "../armem_gui.h"

#include <iostream>

#include <RobotAPI/libraries/armem_gui/instance/sanitize_typename.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace dn = armarx::aron::datanavigator;
namespace tn = armarx::aron::typenavigator;


namespace ArMemGuiTest
{
    struct Fixture
    {
    };

    void test_sanitize(const std::string& in, const std::string& expected)
    {
        const std::string out = armarx::armem::gui::instance::sanitizeTypeName(in);
        BOOST_TEST_CONTEXT("in = '" << in << "'")
        {
            BOOST_CHECK_EQUAL(out, expected);
        }
    }
}



BOOST_FIXTURE_TEST_SUITE(ArMemGuiTest, ArMemGuiTest::Fixture)


BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_int)
{
    test_sanitize(tn::IntNavigator().getName(), "Int");
    test_sanitize(dn::IntNavigator().getName(), "Int");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_float)
{
    test_sanitize(tn::FloatNavigator().getName(), "Float");
    test_sanitize(dn::FloatNavigator().getName(), "Float");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_dict)
{
    tn::DictNavigator dict;
    dict.setAcceptedType(std::make_shared<tn::FloatNavigator>());
    test_sanitize(dict.getName(), "Dict<Float>");

    test_sanitize(dn::DictNavigator().getName(), "Dict");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_list)
{
    tn::ListNavigator list;
    list.setAcceptedType(std::make_shared<tn::LongNavigator>());
    test_sanitize(list.getName(), "List<Long>");

    test_sanitize(dn::ListNavigator().getName(), "List");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_object)
{
    tn::ObjectNavigator obj;
    obj.setObjectName("namespace::MyObjectName");
    test_sanitize(obj.getName(), "MyObjectName    (namespace)");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_tuple)
{
    tn::TupleNavigator type;
    type.addAcceptedType(std::make_shared<tn::IntNavigator>());
    type.addAcceptedType(std::make_shared<tn::FloatNavigator>());
    test_sanitize(type.getName(), "Tuple<Int, Float>");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_pair)
{
    tn::PairNavigator type;
    type.setFirstAcceptedType(std::make_shared<tn::IntNavigator>());
    type.setSecondAcceptedType(std::make_shared<tn::FloatNavigator>());
    test_sanitize(type.getName(), "Pair<Int, Float>");
}

BOOST_AUTO_TEST_CASE(test_sanitizeTypeName_ndarry)
{
    {
        tn::NDArrayNavigator type;
        type.setDimensions({ 3, 2, 1});
        type.setTypename("float");
        test_sanitize(type.getName(), "NDArray");
    }
    {
        dn::NDArrayNavigator data;
        data.setDimensions({ 3, 2, 1});
        test_sanitize(data.getName(), "NDArray<3, 2, 1, >");
    }
}


BOOST_AUTO_TEST_SUITE_END()
