#include "RobotReader.h"

#include <mutex>
#include <optional>

#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/PackagePath.h>

#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem_robot/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot/robot_conversions.h>
#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot_state/aron/JointState.aron.generated.h>


namespace fs = ::std::filesystem;

namespace armarx::armem::robot_state
{

    RobotReader::RobotReader(armem::client::MemoryNameSystem& memoryNameSystem) :
        memoryNameSystem(memoryNameSystem), transformReader(memoryNameSystem)
    {
    }

    void RobotReader::registerPropertyDefinitions(::armarx::PropertyDefinitionsPtr& def)
    {
        transformReader.registerPropertyDefinitions(def);

        def->optional(properties.memoryName, propertyPrefix + "Memory");
        def->optional(properties.descriptionCoreSegment, propertyPrefix + "descriptionSegment");
        def->optional(properties.localizationCoreSegment, propertyPrefix + "localizationSegment");
        def->optional(properties.proprioceptionCoreSegment,
                      propertyPrefix + "proprioceptionSegment");
    }

    void RobotReader::connect()
    {
        transformReader.connect();

        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "RobotReader: Waiting for memory '" << properties.memoryName << "' ...";
        try
        {
            memoryReader = memoryNameSystem.useReader(properties.memoryName);
            ARMARX_IMPORTANT << "RobotReader: Connected to memory '" << properties.memoryName << "'";
        }
        catch (const armem::error::CouldNotResolveMemoryServer& e)
        {
            ARMARX_ERROR << e.what();
            return;
        }
    }

    std::optional<robot::Robot> RobotReader::get(const std::string& name,
            const armem::Time& timestamp)
    {
        const auto description = queryDescription(name, timestamp);

        if (not description)
        {
            ARMARX_ERROR << "Unknown object " << name;
            return std::nullopt;
        }

        return get(*description, timestamp);
    }

    robot::Robot RobotReader::get(const robot::RobotDescription& description,
                                  const armem::Time& timestamp)
    {
        robot::Robot robot{.description = description,
                           .instance    = "", // TODO(fabian.reister):
                           .config      = {}, // will be populated by synchronize
                           .timestamp   = timestamp};

        synchronize(robot, timestamp);

        return robot;
    }

    bool RobotReader::synchronize(robot::Robot& obj, const armem::Time& timestamp)
    {
        auto state = queryState(obj.description, timestamp);

        if (not state) /* c++20 [[unlikely]] */
        {
            ARMARX_WARNING << "Could not synchronize object " << obj.description.name;
            return false;
        }

        obj.config = std::move(*state);
        return true;
    }

    std::optional<robot::RobotDescription>
    RobotReader::queryDescription(const std::string& name, const armem::Time& timestamp)
    {
        // Query all entities from provider.
        armem::client::query::Builder qb;

        // clang-format off
        qb
        .coreSegments().all() // withName(properties.descriptionCoreSegment)
        .providerSegments().all() // .withName(name)
        .entities().all()
        .snapshots().latest(); // TODO(fabian.reister): atTime(timestamp);
        // clang-format on

        ARMARX_INFO << "Lookup query in reader";

        if (not memoryReader)
        {
            ARMARX_WARNING << "Memory reader is null";
            return std::nullopt;
        }

        try
        {
            const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

            ARMARX_INFO << "Lookup result in reader: " << qResult;

            if (not qResult.success) /* c++20 [[unlikely]] */
            {
                return std::nullopt;
            }

            return getRobotDescription(qResult.memory, name);
        }
        catch (...)
        {
            ARMARX_WARNING << "query description failure" << GetHandledExceptionString();
        }

        return std::nullopt;
    }

    std::optional<robot::RobotState>
    RobotReader::queryState(const robot::RobotDescription& description,
                            const armem::Time& timestamp)
    {
        const auto jointMap = queryJointState(description, timestamp);
        if (not jointMap)
        {
            ARMARX_WARNING << "Failed to query joint state";
            return std::nullopt;
        }

        const auto globalPose = queryGlobalPose(description, timestamp);
        if (not globalPose)
        {
            ARMARX_WARNING << "Failed to query global pose";
            return std::nullopt;
        }

        return robot::RobotState
        {
            .timestamp = timestamp, .globalPose = *globalPose, .jointMap = *jointMap};
    }

    std::optional<robot::RobotState::JointMap>
    RobotReader::queryJointState(const robot::RobotDescription& description,
                                 const armem::Time& timestamp) const
    {
        // TODO(fabian.reister): how to deal with multiple providers?

        // Query all entities from provider.
        armem::client::query::Builder qb;

        // clang-format off
        qb
        .coreSegments().withName(properties.proprioceptionCoreSegment)
        .providerSegments().withName(description.name) // agent
        .entities().all() // TODO
        .snapshots().latest();
        // clang-format on

        const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "Lookup result in reader: " << qResult;

        if (not qResult.success) /* c++20 [[unlikely]] */
        {
            ARMARX_WARNING << qResult.errorMessage;
            return std::nullopt;
        }

        return getRobotJointState(qResult.memory, description.name);
    }

    std::optional<robot::RobotState::Pose>
    RobotReader::queryGlobalPose(const robot::RobotDescription& description,
                                 const armem::Time& timestamp) const
    {
        const auto result = transformReader.getGlobalPose(description.name, "root", timestamp);
        if (not result)
        {
            return std::nullopt;
        }

        return result.transform.transform;
    }

    std::optional<robot::RobotState>
    RobotReader::getRobotState(const armarx::armem::wm::Memory& memory,
                               const std::string& name) const
    {
        // clang-format off
        const armem::wm::ProviderSegment& providerSegment = memory
                .getCoreSegment(properties.proprioceptionCoreSegment)
                .getProviderSegment(name);

        // TODO entitiesToRobotState()

        if (providerSegment.empty())
        {
            ARMARX_WARNING << "No entity found";
            return std::nullopt;
        }

        // clang-format on

        const armem::wm::EntityInstance* instance = nullptr;
        providerSegment.forEachInstance([&instance](const wm::EntityInstance & i)
        {
            instance = &i;
            return false;  // break
        });
        if (!instance)
        {
            ARMARX_WARNING << "No entity snapshots found";
            return std::nullopt;
        }

        // Here, we access the RobotUnit streaming data stored in the proprioception segment.
        return robot::convertRobotState(*instance);
    }

    // FIXME remove this, use armem/util/util.h
    template <typename AronClass>
    std::optional<AronClass> tryCast(const wm::EntityInstance& item)
    {
        static_assert(std::is_base_of<armarx::aron::cppserializer::AronCppClass,
                      AronClass>::value);

        try
        {
            AronClass t;
            t.fromAron(item.data());
            return t;
        }
        catch (const armarx::aron::error::AronException&)
        {
            return std::nullopt;
        }
    }

    std::optional<robot::RobotState::JointMap>
    RobotReader::getRobotJointState(const armarx::armem::wm::Memory& memory,
                                    const std::string& name) const
    {

        robot::RobotState::JointMap jointMap;

        // clang-format off
        const armem::wm::CoreSegment& coreSegment = memory
                .getCoreSegment(properties.proprioceptionCoreSegment);
        // clang-format on

        coreSegment.forEachEntity([&jointMap](const wm::Entity & entity)
        {
            const auto& entityInstance = entity.getLatestSnapshot().getInstance(0);

            const auto jointState = tryCast<::armarx::armem::arondto::JointState>(entityInstance);
            if (not jointState)
            {
                // ARMARX_WARNING << "Could not convert entity instance to 'JointState'";
                return;
            }

            jointMap.emplace(jointState->name, jointState->position);
        });

        if (jointMap.empty())
        {
            return std::nullopt;
        }

        return jointMap;
    }

    std::optional<robot::RobotDescription>
    RobotReader::getRobotDescription(const armarx::armem::wm::Memory& memory,
                                     const std::string& name) const
    {
        // clang-format off
        const armem::wm::ProviderSegment& providerSegment = memory
                .getCoreSegment(properties.descriptionCoreSegment)
                .getProviderSegment(name); // TODO(fabian.reister): all
        // clang-format on

        const armem::wm::EntityInstance* instance = nullptr;
        providerSegment.forEachInstance([&instance](const wm::EntityInstance & i)
        {
            instance = &i;
        });
        if (!instance)
        {
            ARMARX_WARNING << "No entity snapshots found";
            return std::nullopt;
        }

        return robot::convertRobotDescription(*instance);
    }

} // namespace armarx::armem::robot_state
