#include "Segment.h"

#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/libraries/DebugObserverHelper/DebugObserverHelper.h>
#include <ArmarXCore/observers/ObserverObjectFactories.h>

#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem_robot_state/aron/Proprioception.aron.generated.h>


namespace armarx::armem::server::robot_state::proprioception
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        Base(memoryToIceAdapter, "Proprioception", arondto::Proprioception::toAronType(), 1024)
    {
    }

    Segment::~Segment() = default;


    void Segment::onConnect(RobotUnitInterfacePrx robotUnitPrx)
    {
        this->robotUnit = robotUnitPrx;

        std::string providerSegmentName = "Robot";
        KinematicUnitInterfacePrx kinematicUnit = robotUnit->getKinematicUnit();
        if (kinematicUnit)
        {
            providerSegmentName = kinematicUnit->getRobotName();
        }
        else
        {
            ARMARX_WARNING << "Robot unit '" << robotUnit->ice_getIdentity().name << "' does not have a kinematic unit."
                           << "\n Falling back to provider segment name '" << providerSegmentName << "'.";
        }
        this->robotUnitProviderID = segment->id().withProviderSegmentName(providerSegmentName);
    }



    RobotJointPositionMap Segment::getRobotJointPositionsLocking(
        const armem::Time& timestamp,
        DebugObserverHelper* debugObserver) const
    {
        return doLocked([this, &timestamp, &debugObserver]()
        {
            return getRobotJointPositions(timestamp, debugObserver);
        });
    }


    static
    aron::datanavigator::DictNavigatorPtr
    getDictElement(const aron::datanavigator::DictNavigator& dict, const std::string& key)
    {
        if (dict.hasElement(key))
        {
            return aron::datanavigator::DictNavigator::DynamicCastAndCheck(dict.getElement(key));
        }
        return nullptr;
    }


    RobotJointPositionMap
    Segment::getRobotJointPositions(
        const armem::Time& timestamp,
        DebugObserverHelper* debugObserver) const
    {
        namespace adn = aron::datanavigator;
        ARMARX_CHECK_NOT_NULL(segment);

        RobotJointPositionMap jointMap;
        int i = 0;

        Duration tFindData = Duration::milliSeconds(0), tReadJointPositions = Duration::milliSeconds(0);
        TIMING_START(tProcessEntities)
        segment->forEachEntity([&](const wm::Entity & entity)
        {
            adn::DictNavigatorPtr data;
            {
                TIMING_START(_tFindData)

                const wm::EntitySnapshot* snapshot = entity.findLatestSnapshotBeforeOrAt(timestamp);
                if (not snapshot)
                {
                    // Got no snapshot <= timestamp. Take latest instead (if present).
                    snapshot = entity.findLatestSnapshot();
                }
                if (snapshot)
                {
                    data = snapshot->findInstanceData();
                }

                TIMING_END_COMMENT_STREAM(_tFindData, "tFindData " + std::to_string(i), ARMARX_DEBUG)
                tFindData += _tFindData;
            }
            if (data)
            {
                TIMING_START(_tReadJointPositions)

                jointMap.emplace(entity.id().providerSegmentName, readJointPositions(*data));

                TIMING_END_COMMENT_STREAM(_tReadJointPositions, "tReadJointPositions " + std::to_string(i), ARMARX_DEBUG)
                tReadJointPositions += _tReadJointPositions;
            }
            ++i;
        });
        TIMING_END_STREAM(tProcessEntities, ARMARX_DEBUG)

        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield(dp + "t 1.1 Process Entities (ms)", tProcessEntities.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(dp + "t 1.1.1 FindData (ms)", tFindData.toMilliSecondsDouble());
            debugObserver->setDebugObserverDatafield(dp + "t 1.1.2 ReadJointPositions (ms)", tReadJointPositions.toMilliSecondsDouble());
        }

        return jointMap;
    }


    const armem::MemoryID& Segment::getRobotUnitProviderID() const
    {
        return robotUnitProviderID;
    }


    std::map<std::string, float>
    Segment::readJointPositions(const wm::EntityInstanceData& data)
    {
        namespace adn = aron::datanavigator;

        // Just get what we need without casting the whole data.
        std::map<std::string, float> jointPositions;
        if (adn::DictNavigatorPtr joints = getDictElement(data, "joints"))
        {
            if (adn::DictNavigatorPtr jointsPosition = getDictElement(*joints, "position"))
            {
                for (const auto& [name, value] : jointsPosition->getElements())
                {
                    const float jointPosition = adn::FloatNavigator::DynamicCastAndCheck(value)->getValue();
                    jointPositions[name] = jointPosition;
                }
            }
        }
        return jointPositions;
    }

}  // namespace armarx::armem::server::robot_state::proprioception
