/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Dict.h"


namespace armarx::aron::cppserializer::serializer
{
    // constructors
    DictSerializer::DictSerializer(const typenavigator::DictNavigatorPtr& e) :
        detail::ContainerSerializerBase<typenavigator::DictNavigator, DictSerializer>("std::map<std::string, " + FromAronTypeNaviagtorPtr(e->getAcceptedType())->getFullCppTypename() + ">",
                simox::meta::get_type_name(typeid(data::AronDict)),
                simox::meta::get_type_name(typeid(type::AronDict)), e)
    {
    }

    // virtual implementations
    CppBlockPtr DictSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        std::string resolved_accessor = ResolveMaybeAccessor(accessor, typenavigator);

        block_if_data->addLine(accessor + nextEl() + "clear();");
        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr DictSerializer::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine(accessor + " = " + this->getFullTypenameGenerator() + "();");
        return ResolveMaybeResetHardBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr DictSerializer::getWriteTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartDict({" + MAYBE_TO_STRING(typenavigator->getMaybe()) + "}); // of " + accessor);

        auto type_s = FromAronTypeNaviagtorPtr(typenavigator->getAcceptedType());
        CppBlockPtr b2 = type_s->getWriteTypeBlock(type_s->getFullCppTypename());
        b->appendBlock(b2);
        b->addLine("w.writeEndDict(); // of " + accessor);
        return b;
    }

    CppBlockPtr DictSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = std::make_shared<CppBlock>();
        block_if_data->addLine("w.writeStartDict(); // of " + accessor);

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_iterator_key = escaped_accessor + DICT_KEY_ACCESSOR;
        std::string accessor_iterator_val = escaped_accessor + DICT_VALUE_ACCESSOR;

        std::string resolved_accessor = ResolveMaybeAccessor(accessor, typenavigator);

        block_if_data->addLine("for (const auto& [" + accessor_iterator_key + ", " + accessor_iterator_val + "] : " + resolved_accessor + ") ");

        auto type_s = FromAronTypeNaviagtorPtr(typenavigator->getAcceptedType());
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine("w.writeKey(" + accessor_iterator_key + ");");
        b2->appendBlock(type_s->getWriteBlock(accessor_iterator_val));
        block_if_data->addBlock(b2);
        block_if_data->addLine("w.writeEndDict(); // of " + accessor);

        return ResolveMaybeWriteBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr DictSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_iterator = escaped_accessor + DICT_ITERATOR_ACCESSOR;
        std::string accessor_iterator_key = escaped_accessor + DICT_KEY_ACCESSOR;

        auto type_s = FromAronTypeNaviagtorPtr(typenavigator->getAcceptedType());

        block_if_data->addLine("while(!r.readEndDict()) // of " + accessor);
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());

        b2->addLine("std::string " + accessor_iterator_key + " = r.readKey(); // in dict " + accessor);
        b2->addLine(type_s->getFullCppTypename() + " " + accessor_iterator + ";");
        b2->appendBlock(type_s->getReadBlock(accessor_iterator));
        b2->addLine(accessor + nextEl() + "insert({" + accessor_iterator_key + ", " + accessor_iterator + "});");

        block_if_data->addBlock(b2);
        return ResolveMaybeReadBlock(accessor, "r.readStartDict()", block_if_data, typenavigator);
    }

    CppBlockPtr DictSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
        block_if_data->addLineAsBlock("return false;");
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }
}

