/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Tuple.h"


namespace armarx::aron::cppserializer::serializer
{
    TupleSerializer::TupleSerializer(const typenavigator::TupleNavigatorPtr& e) :
        detail::ContainerSerializerBase<typenavigator::TupleNavigator, TupleSerializer>("std::tuple<" + simox::alg::join(ExtractCppTypenames(e->getAcceptedTypes()), ", ") + ">",
                simox::meta::get_type_name(typeid(data::AronList)),
                simox::meta::get_type_name(typeid(type::AronTuple)), e)
    {
    }

    CppBlockPtr TupleSerializer::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine(accessor + " = " + this->getFullTypenameGenerator() + "();");
        return ResolveMaybeResetHardBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr TupleSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        std::string resolved_accessor = ResolveMaybeAccessor(accessor, typenavigator);

        unsigned int i = 0;
        for (const auto& child : typenavigator->getAcceptedTypes())
        {
            auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getResetSoftBlock("std::get<" + std::to_string(i++) + ">(" + resolved_accessor + ")");
            block_if_data->appendBlock(b2);
        }
        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr TupleSerializer::getWriteTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartTuple({" + MAYBE_TO_STRING(typenavigator->getMaybe()) + "}); // of " + accessor);

        std::string escaped_accessor = EscapeAccessor(accessor);
        unsigned int i = 0;
        for (const auto& type : typenavigator->getAcceptedTypes())
        {
            std::string accessor_iterator = escaped_accessor + TUPLE_ITERATOR_ACCESSOR + std::to_string(i++);
            auto type_s = FromAronTypeNaviagtorPtr(type);
            b->addLine("w.writeKey(\"" + std::to_string(i) + "\");");
            CppBlockPtr b2 = type_s->getWriteTypeBlock(accessor_iterator);
            b->appendBlock(b2);
        }
        b->addLine("w.writeEndTuple(); // of " + accessor);
        return b;
    }

    CppBlockPtr TupleSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("w.writeStartList(); // of " + accessor);

        std::string resolved_accessor = ResolveMaybeAccessor(accessor, typenavigator);

        unsigned int i = 0;
        for (const auto& type : typenavigator->getAcceptedTypes())
        {
            auto type_s = FromAronTypeNaviagtorPtr(type);
            CppBlockPtr b2 = type_s->getWriteBlock("std::get<" + std::to_string(i++) + ">(" + resolved_accessor + ")");
            block_if_data->appendBlock(b2);
        }
        block_if_data->addLine("w.writeEndList(); // of " + accessor);

        return ResolveMaybeWriteBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr TupleSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());

        std::string resolved_accessor = ResolveMaybeAccessor(accessor, typenavigator);

        unsigned int i = 0;
        for (const auto& type : typenavigator->getAcceptedTypes())
        {
            auto type_s = FromAronTypeNaviagtorPtr(type);
            CppBlockPtr b2 = type_s->getReadBlock("std::get<" + std::to_string(i++) + ">(" + resolved_accessor + ")");
            block_if_data->appendBlock(b2);
        }
        block_if_data->addLine("r.readEndList(); // of " + accessor);
        return ResolveMaybeReadBlock(accessor, "r.readStartList()", block_if_data, typenavigator);
    }

    CppBlockPtr TupleSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
        block_if_data->addLineAsBlock("return false;");
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }
}

