/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Data.h"

#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

#include <SimoxUtility/algorithm/string/string_tools.h>

namespace armarx::aron::xmltypereader
{

    void Data::EnforceAttribute(const RapidXmlReaderNode& node, const std::string& att)
    {
        if (!HasAttribute(node, att))
        {
            throw error::StringNotValidException("XMLReaderData", "EnforceAttribute", "A <" + node.name() + ">-tag does not have the correct attribute", att);
        }
    }

    bool Data::HasAttribute(const RapidXmlReaderNode& node, const std::string& att)
    {
        return node.has_attribute(att.c_str());
    }

    std::string Data::GetAttribute(const RapidXmlReaderNode& node, const std::string& att)
    {
        EnforceAttribute(node, att);
        return node.attribute_value(att.c_str());
    }

    std::string Data::GetAttributeWithDefault(const armarx::RapidXmlReaderNode& node, const std::string& att, const std::string def)
    {
        if (!(HasAttribute(node, att)))
        {
            return def;
        }
        return node.attribute_value(att.c_str());
    }

    bool Data::AttributeIsTrue(const armarx::RapidXmlReaderNode& node, const std::string& att)
    {
        if (HasAttribute(node, att))
        {
            std::string v = simox::alg::to_lower(node.attribute_value(att.c_str()));
            if (v == "1" or v == "true" or v == "wahr" or v == "yes" or v == "ja" or v == "")
            {
                return true;
            }
        }
        return false;
    }

    bool Data::HasTagName(const armarx::RapidXmlReaderNode& node, const std::string& name)
    {
        return (simox::alg::to_lower(name) == simox::alg::to_lower(node.name()));
    }

    void Data::EnforceTagName(const armarx::RapidXmlReaderNode& node, const std::string& name)
    {
        if (!(HasTagName(node, name)))
        {
            throw error::StringNotValidException("XMLReaderData", "EnforceTagName", "The node <" + node.name() + "> has the wrong tag", name);
        }
    }

    std::string Data::GetTagName(const armarx::RapidXmlReaderNode& node)
    {
        return simox::alg::to_lower(node.name());
    }

    bool Data::CheckMinChildSize(const armarx::RapidXmlReaderNode& node, const size_t size)
    {
        std::vector<RapidXmlReaderNode> children = node.nodes();
        return children.size() >= size;
    }

    bool Data::CheckMaxChildSize(const armarx::RapidXmlReaderNode& node, const size_t size)
    {
        std::vector<RapidXmlReaderNode> children = node.nodes();
        return children.size() <= size;
    }

    bool Data::CheckExactChildSize(const armarx::RapidXmlReaderNode& node, const size_t size)
    {
        std::vector<RapidXmlReaderNode> children = node.nodes();
        return children.size() == size;
    }

    void Data::EnforceChildSize(const armarx::RapidXmlReaderNode& node, const size_t size)
    {
        if (!Data::CheckExactChildSize(node, size))
        {
            throw error::SizeNotValidException("XMLReaderData", "EnforceChildSize", "The node <" + node.name() + "> has the wrong number of children", node.nodes().size(), size);
        }
    }

}
