/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL


// Header
#include "Reader.h"

// ArmarX
#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>
#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>

#include <RobotAPI/libraries/aron/core/codegenerator/typeReader/xml/Data.h>
#include <RobotAPI/libraries/aron/core/navigator/type/NavigatorFactory.h>


namespace fs = std::filesystem;

namespace armarx::aron::xmltypereader
{
    void Reader::parseFile(const std::string& _filename)
    {
        std::string filename = _filename;
        // Handle C++ style includes like "<path/to/file>".
        if (!filename.empty() && filename.front() == '<' && filename.back() == '>')
        {
            filename = filename.substr(1, filename.size() - 2);
        }
        parseFile(std::filesystem::path(filename));
    }

    void Reader::parseFile(const std::filesystem::path& _file)
    {
        fs::path file = _file;

        RapidXmlReaderPtr reader = RapidXmlReader::FromFile(file.string());
        parse(reader, _file);
    }

    // private method reading nodes
    void Reader::parse(const RapidXmlReaderPtr& reader, const std::filesystem::path& filePath)
    {
        RapidXmlReaderNode root = reader->getRoot();

        // Check Header
        Data::EnforceTagName(root, Data::TYPE_DEFINITION_TAG);

        Data::CheckMaxChildSize(root, 4);
        std::vector<RapidXmlReaderNode> children = root.nodes();

        int cpp_includes_index = -1;
        int include_aron_file_index = -1;
        int generate_types_index = -1;

        int index = 0;
        for (const auto& child : children)
        {
            if (Data::HasTagName(child, Data::CODE_INCLUDES_TAG))
            {
                if (cpp_includes_index != -1)
                {
                    throw error::AronException("XMLReader", "parse", "Already found an include tag. Please check the xml file.");
                }
                else
                {
                    cpp_includes_index = index;
                }
            }
            else if (Data::HasTagName(child, Data::INCLUDES_TAG))
            {
                if (include_aron_file_index != -1)
                {
                    throw error::AronException("XMLReader", "parse", "Already found an use type tag. Please check the xml file.");
                }
                else
                {
                    include_aron_file_index = index;
                }
            }
            else if (Data::HasTagName(child, Data::GENERATE_TYPES_TAG))
            {
                if (generate_types_index != -1)
                {
                    throw error::AronException("XMLReader", "parse", "Already found an generate type tag. Please check the xml file.");
                }
                else
                {
                    generate_types_index = index;
                }
            }
            else
            {
                throw error::StringNotValidException("XMLReader", "parse", "Found an unexpected xml tag", child.name());
            }
            index++;
        }

        // 1. Check includes
        if (cpp_includes_index != -1)
        {
            std::vector<RapidXmlReaderNode> includes = children[cpp_includes_index].nodes();
            for (const auto& include : includes)
            {
                this->codeIncludes.push_back(readCodeInclude(include, filePath));
            }
        }

        // 2. Check AronIncludes
        if (include_aron_file_index != -1)
        {
            for (const auto& aronInclude : children[include_aron_file_index].nodes())
            {
                // right now unused
                this->aronIncludes.push_back(readAronInclude(aronInclude, filePath));
            }
        }

        // 3. Check GenerateTypes
        if (generate_types_index != -1)
        {
            for (const auto& generateType : children[generate_types_index].nodes())
            {
                // ugly workaround
                if (Data::HasTagName(generateType, Data::GENERATE_OBJECT_TAG))
                {
                    const auto nav = readGenerateObject(generateType);
                    const auto infos = factory.AllPublicObjects;
                    generateObjects.push_back(infos.at(simox::alg::to_lower(nav->getObjectName())));
                    continue;
                }
                if (Data::HasTagName(generateType, Data::GENERATE_INT_ENUM_TAG))
                {
                    const auto nav = readGenerateIntEnum(generateType);
                    const auto infos = factory.AllPublicIntEnums;
                    generateIntEnums.push_back(infos.at(simox::alg::to_lower(nav->getEnumName())));
                    continue;
                }
                throw error::StringNotValidException("XMLReader", "parse", "Could not find a valid tag inside generatetypes", generateType.name());

            }
        }
    }

    std::string Reader::readCodeInclude(const RapidXmlReaderNode& node, const std::filesystem::path& filePath)
    {
        Data::EnforceTagName(node, Data::INCLUDE_TAG);
        Data::EnforceAttribute(node, Data::INCLUDE_ATTRIBUTE_NAME);
        const std::string include = node.attribute_value(Data::INCLUDE_ATTRIBUTE_NAME);
        return include;
    }

    std::string Reader::readAronInclude(const RapidXmlReaderNode& node, const std::filesystem::path& filePath)
    {
        Data::EnforceTagName(node, Data::INCLUDE_TAG);
        std::string specifiedPath = Data::GetAttribute(node, Data::INCLUDE_ATTRIBUTE_NAME);
        specifiedPath = simox::alg::replace_all(specifiedPath, "<", "");
        specifiedPath = simox::alg::replace_all(specifiedPath, ">", "");
        const std::filesystem::path xmlincludepath(specifiedPath);

        // add logic to allow relative paths:
        std::filesystem::path resolved_absolute_path(xmlincludepath);
        std::filesystem::path resolved_relative_path(xmlincludepath);

        if (!xmlincludepath.empty() && xmlincludepath.is_relative())
        {
            if (std::optional<fs::path> resolvedPackagePath = resolveRelativePackagePath(xmlincludepath); resolvedPackagePath.has_value())
            {
                // check if relative path belongs to a package
                resolved_absolute_path = resolvedPackagePath.value();
                resolved_relative_path = xmlincludepath;
            }
            else
            {
                // perhaps the relative path is relative from file position?
                // TODO: FixMe
                resolved_absolute_path = filePath.parent_path() / xmlincludepath;
                resolved_relative_path = filePath.parent_path() / xmlincludepath;

                if (!fs::is_regular_file(resolved_absolute_path))
                {
                    throw error::AronException("Reader", "readAronInclude", "Could not find an aron XML file. Last path tried was: " + resolved_absolute_path.string());
                }
            }
        }
        // else path is absolute

        // parse parent xml file and add objects to alreday known
        Reader anotherReader;
        anotherReader.parseFile(resolved_absolute_path);

        if (Data::HasAttribute(node, Data::AUTO_CODE_INCLUDE))
        {
            std::string codeinclude = simox::alg::replace_last(resolved_relative_path, ".xml", ".aron.generated.h");
            this->codeIncludes.push_back("<" + codeinclude + ">");
        }

        return resolved_absolute_path.string();
    }

    typenavigator::ObjectNavigatorPtr Reader::readGenerateObject(const RapidXmlReaderNode& node) const
    {
        Data::EnforceTagName(node, Data::GENERATE_OBJECT_TAG);
        return typenavigator::ObjectNavigator::DynamicCastAndCheck(factory.create(node, Path()));
    }

    typenavigator::IntEnumNavigatorPtr Reader::readGenerateIntEnum(const RapidXmlReaderNode& node) const
    {
        Data::EnforceTagName(node, Data::GENERATE_INT_ENUM_TAG);
        return typenavigator::IntEnumNavigator::DynamicCastAndCheck(factory.create(node, Path()));
    }


    std::optional<fs::path> Reader::resolveRelativePackagePath(const fs::path& path)
    {
        const std::string package = *path.begin();
        armarx::CMakePackageFinder finder(package);
        if (finder.packageFound())
        {
            for (const std::string& includePath : finder.getIncludePathList())
            {
                fs::path absPath = includePath / path;
                if (fs::is_regular_file(absPath))
                {
                    // path is valid
                    return absPath;
                }
            }
            return std::nullopt;
        }
        else
        {
            return std::nullopt;
        }
    }

}


