/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorWriter.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx::aron::typeIO::writer
{
    // generate Path info
    Path NavigatorWriter::generatePath()
    {
        Path path;
        if (!wroteInitialStartObject)
        {
            wroteInitialStartObject = true;
        }
        else
        {
            auto current = stack.top();
            path = Path(current->getElement()->getPath(), current->toAccessor());
        }
        return path;
    }

    // interface
    void NavigatorWriter::writeStartObject(const WriteStartObjectInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ObjectNavigator>(path);
        type->setObjectName(o.name);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndObject()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartDict(const WriteStartDictInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::DictNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartList(const WriteStartListInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ListNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartTuple(const WriteStartTupleInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::TupleNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndTuple()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartPair(const WriteStartPairInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::PairNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndPair()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeEigenMatrix(const WriteEigenMatrixInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::EigenMatrixNavigator>(path);
        type->setRows(o.rows);
        type->setCols(o.cols);
        type->setTypename(o.type);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeEigenQuaternion(const WriteEigenQuaternionInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::EigenQuaternionNavigator>(path);
        type->setTypename(o.type);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeIVTCByteImage(const WriteIVTCByteImageInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::IVTCByteImageNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeOpenCVMat(const WriteOpenCVMatInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::OpenCVMatNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writePCLPointCloud(const WritePCLPointCloudInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::PCLPointCloudNavigator>(path);
        type->setTypename(o.type);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writePosition(const WritePositionInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::PositionNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeOrientation(const WriteOrientationInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::OrientationNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writePose(const WritePoseInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::PoseNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeInt(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::IntNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeLong(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::LongNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeFloat(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::FloatNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeDouble(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::DoubleNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeString(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::StringNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeBool(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::BoolNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeTime(const WritePrimitiveInput& o)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<typenavigator::TimeNavigator>(path);
        type->setMaybe(o.maybe);
        token->addElement(type);
    }

    void NavigatorWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    void NavigatorWriter::writeExtends()
    {
        if (stack.size() < 2)
        {
            throw error::SizeNotValidException("NavigatorWriter", "writeExtends", "The stacksize must at least be 2 (one for base, one for derived)", stack.size(), 2);
        }

        auto base = stack.top();
        stack.pop();
        auto derived = stack.top();

        auto base_obj = typenavigator::ObjectNavigator::DynamicCastAndCheck(base->getElement());
        auto derived_obj = typenavigator::ObjectNavigator::DynamicCastAndCheck(derived->getElement());
        derived_obj->setExtends(base_obj);
    }
}
