/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::aron
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2019
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::aron

#define ARMARX_BOOST_TEST

// STD/STL
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <numeric>

// IVT
#include <Image/ByteImage.h>

// Eigen
#include <Eigen/Core>

// OpenCV
#include <opencv2/core/core.hpp>

// PCL
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>

// Test
#include <RobotAPI/Test.h>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>
#include <RobotAPI/libraries/aron/core/Exception.h>

// Aron
#include <RobotAPI/libraries/aron/core/Debug.h>
#include <RobotAPI/libraries/aron/core/Randomizer.h>

#include <RobotAPI/libraries/aron/core/io/dataIO/reader/navigator/NavigatorReader.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/nlohmannJSON/NlohmannJSONReader.h>

#include <RobotAPI/libraries/aron/core/io/dataIO/writer/navigator/NavigatorWriter.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/writer/nlohmannJSON/NlohmannJSONWriter.h>

#include <RobotAPI/libraries/aron/core/io/typeIO/reader/navigator/NavigatorReader.h>
#include <RobotAPI/libraries/aron/core/io/typeIO/reader/nlohmannJSON/NlohmannJSONReader.h>

#include <RobotAPI/libraries/aron/core/io/typeIO/writer/navigator/NavigatorWriter.h>
#include <RobotAPI/libraries/aron/core/io/typeIO/writer/nlohmannJSON/NlohmannJSONWriter.h>

// Files without code generation
//#include "aronDataWithoutCodeGeneration.h"

// Generated File
#include <RobotAPI/libraries/aron/core/test/aron/ListTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/DictTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PrimitiveTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/ObjectTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/IVTCByteImageTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EigenMatrixTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EigenQuaternionTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/OpenCVMatTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PCLPointCloudTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PositionTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/OrientationTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/PoseTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/EnumTest.aron.generated.h>
#include <RobotAPI/libraries/aron/core/test/aron/OptionalTest.aron.generated.h>

using namespace armarx;
using namespace aron;

template <typename T>
void runTestWithInstances(T& k1, T& k2)
{
    // assumes not nullptrs as k1 and k2. If you have a maybe type then make sure that it is set properly

    int echo_debug_lv = 1;

    Randomizer r;

    {
        if (echo_debug_lv)
        {
            std::cout << "\t getting type 1" << std::endl;
        }
        typenavigator::ObjectNavigatorPtr k1_type_nav = k1.toAronType();
        type::AronObjectPtr k1_type = k1_type_nav->toAronObjectPtr();
        BOOST_CHECK_EQUAL((bool) k1_type, true);

        armarx::aron::typeIO::writer::NlohmannJSONWriter json_writer_k1;
        k1.writeType(json_writer_k1);
        nlohmann::json k1_type_json = json_writer_k1.getResult();
        BOOST_CHECK_EQUAL(k1_type_json.is_object(), true);

        if (echo_debug_lv)
        {
            std::cout << "\t getting type 2" << std::endl;
        }
        typenavigator::ObjectNavigatorPtr k2_type_nav = k2.toAronType();
        type::AronObjectPtr k2_type = k2_type_nav->toAronObjectPtr();
        BOOST_CHECK_EQUAL((bool) k2_type, true);

        armarx::aron::typeIO::writer::NlohmannJSONWriter json_writer_k2;
        k2.writeType(json_writer_k2);
        nlohmann::json k2_type_json = json_writer_k2.getResult();
        BOOST_CHECK_EQUAL(k2_type_json.is_object(), true);
        BOOST_CHECK_EQUAL(k1_type_json.dump(2), k2_type_json.dump(2));
    }

    {
        typenavigator::ObjectNavigatorPtr k1_type_nav = k1.toAronType();
        datanavigator::DictNavigatorPtr k1_aron_nav = k1.toAron();

        if (echo_debug_lv)
        {
            std::cout << "\t initialize aron 1 randomly" << std::endl;
        }
        r.initializeRandomly(k1_aron_nav, k1_type_nav);

        if (echo_debug_lv)
        {
            std::cout << "\t getting aron 1" << std::endl;
        }
        data::AronDictPtr k1_aron = k1_aron_nav->toAronDictPtr();
        BOOST_CHECK_EQUAL((bool) k1_aron, true);

        if (echo_debug_lv >= 2)
        {
            std::cout << "K1 Aron:" << std::endl;
            std::cout << armarx::aron::Debug::AronDataPtrToString(k1_aron) << std::endl;
            std::cout << "K2 Aron:" << std::endl;
            std::cout << armarx::aron::Debug::AronDataPtrToString(k2.toAron()->toAronDictPtr()) << std::endl;
        }

        if (echo_debug_lv)
        {
            std::cout << "\t setting aron 2 from aron 1" << std::endl;
        }
        k2.fromAron(k1_aron);

        if (echo_debug_lv)
        {
            std::cout << "\t getting aron 2" << std::endl;
        }
        datanavigator::DictNavigatorPtr k2_aron_nav = k2.toAron();
        BOOST_CHECK_EQUAL((*k1_aron_nav == *k2_aron_nav), true);

        data::AronDictPtr k2_aron = k2_aron_nav->toAronDictPtr();
        BOOST_CHECK_EQUAL((bool) k2_aron, true);

        if (echo_debug_lv)
        {
            std::cout << "\t setting aron 1 from aron 2 and check for equality" << std::endl;
        }
        k1.fromAron(k2_aron_nav);

        if (echo_debug_lv >= 2)
        {
            std::cout << "K1 Aron:" << std::endl;
            std::cout << armarx::aron::Debug::AronDataPtrToString(k1.toAron()->toAronDictPtr()) << std::endl;
            std::cout << "K2 Aron:" << std::endl;
            std::cout << armarx::aron::Debug::AronDataPtrToString(k2.toAron()->toAronDictPtr()) << std::endl;
        }
        BOOST_CHECK_EQUAL((k1 == k2), true);

        datanavigator::DictNavigatorPtr k1_aron_nav_again = k1.toAron();
        BOOST_CHECK_EQUAL((*k1_aron_nav == *k1_aron_nav_again), true);
    }

    {
        typenavigator::ObjectNavigatorPtr k1_type_nav = k1.toAronType();
        datanavigator::DictNavigatorPtr k1_aron_nav = k1.toAron();

        r.initializeRandomly(k1_aron_nav, k1_type_nav);

        data::AronDictPtr k1_aron = k1_aron_nav->toAronDictPtr();
        k1.fromAron(k1_aron);

        if (echo_debug_lv)
        {
            std::cout << "\t check JSON export of k and k2 for equality" << std::endl;
        }
        armarx::aron::dataIO::writer::NlohmannJSONWriter json_writer_for_k1;
        armarx::aron::dataIO::writer::NlohmannJSONWriter json_writer_for_k2;

        k1.write(json_writer_for_k1);
        nlohmann::json k1_aron_json = json_writer_for_k1.getResult();
        std::string k1_aron_json_str = k1_aron_json.dump(4);

        armarx::aron::dataIO::writer::NlohmannJSONWriter direct_json_writer_for_k1;
        armarx::aron::dataIO::Visitor::VisitAndSetup(direct_json_writer_for_k1, k1_aron);
        nlohmann::json direct_k1_aron_json = direct_json_writer_for_k1.getResult();
        std::string direct_k1_aron_json_str = direct_k1_aron_json.dump(4);

        if (echo_debug_lv >= 2)
        {
            std::cout << "\t K1 as json: " << std::endl << k1_aron_json_str << std::endl;
            std::cout << "\t K1 as direct json: " << std::endl << direct_k1_aron_json_str << std::endl;
        }
        BOOST_CHECK_EQUAL((k1_aron_json_str == direct_k1_aron_json_str), true);

        armarx::aron::dataIO::reader::NlohmannJSONReader json_reader_for_k2(k1_aron_json);

        k2.read(json_reader_for_k2);
        k2.write(json_writer_for_k2);

        nlohmann::json k2_aron_json = json_writer_for_k2.getResult();
        std::string k2_aron_json_str = k2_aron_json.dump(4);

        if (echo_debug_lv >= 2)
        {
            std::cout << "\t K1 as json: " << std::endl << k1_aron_json_str << std::endl;
            std::cout << "\t K2 as json: " << std::endl << k2_aron_json_str << std::endl;
        }
        BOOST_CHECK_EQUAL((k1_aron_json_str == k2_aron_json_str), true);
    }
}

BOOST_AUTO_TEST_CASE(AronListTest)
{
    std::cout << "Running List test" << std::endl;
    ListTest l;
    ListTest l2;
    runTestWithInstances<ListTest>(l, l2);
}

BOOST_AUTO_TEST_CASE(AronDictTest)
{
    std::cout << "Running Dict test" << std::endl;
    DictTest d;
    DictTest d2;
    runTestWithInstances<DictTest>(d, d2);
}

BOOST_AUTO_TEST_CASE(AronPrimitiveTest)
{
    std::cout << "Running Primitive test" << std::endl;
    PrimitiveTest p;
    PrimitiveTest p2;
    runTestWithInstances<PrimitiveTest>(p, p2);
}

BOOST_AUTO_TEST_CASE(AronObjectTest)
{
    std::cout << "Running Object test" << std::endl;
    ObjectTest1 o1;
    ObjectTest1 o12;
    runTestWithInstances<ObjectTest1>(o1, o12);

    ObjectTest2 o2;
    ObjectTest2 o22;
    runTestWithInstances<ObjectTest2>(o2, o22);
}

BOOST_AUTO_TEST_CASE(AronImageTest)
{
    std::cout << "Running Image test" << std::endl;
    IVTCByteImageTest ii;
    IVTCByteImageTest ii2;

    ii.the_grayscale_image = std::make_shared<CByteImage>();
    ii2.the_grayscale_image = std::make_shared<CByteImage>();

    ii.the_rgb_image = std::make_shared<CByteImage>();
    ii2.the_rgb_image = std::make_shared<CByteImage>();

    ii.the_grayscale_image->Set(5, 5, CByteImage::eGrayScale);
    ii2.the_grayscale_image->Set(7, 7, CByteImage::eGrayScale);

    ii.the_rgb_image->Set(5, 5, CByteImage::eRGB24);
    ii2.the_rgb_image->Set(7, 7, CByteImage::eRGB24);

    runTestWithInstances<IVTCByteImageTest>(ii, ii2);
}

BOOST_AUTO_TEST_CASE(AronEigenMatrixTest)
{
    // Eigen may cause problems with dimensions > 145
    std::cout << "Running EigenMatrix test" << std::endl;
    EigenMatrixTest em;
    EigenMatrixTest em2;
    runTestWithInstances<EigenMatrixTest>(em, em2);
}

BOOST_AUTO_TEST_CASE(AronOpenCVTest)
{
    std::cout << "Running OpenCVMat test" << std::endl;
    OpenCVMatTest ocv;
    OpenCVMatTest ocv2;

    ocv.the_2d_opencv_matrix = cv::Mat(2, 2, CV_64F);
    ocv.the_3d_opencv_matrix = cv::Mat(std::vector<int>({2, 2, 2}), CV_32F);
    ocv.the_4d_opencv_matrix = cv::Mat(std::vector<int>({2, 2, 2, 2}), CV_16S);
    ocv.the_5d_opencv_matrix = cv::Mat(std::vector<int>({2, 2, 2, 2, 2}), CV_8U);

    runTestWithInstances<OpenCVMatTest>(ocv, ocv2);
}

BOOST_AUTO_TEST_CASE(AronPCLPointCloudTest)
{
    std::cout << "Running PCLPointCloud test" << std::endl;
    PointCloudTest pc;
    PointCloudTest pc2;

    pc.the_xyzrgb_pcl_pointcloud = pcl::PointCloud<pcl::PointXYZRGB>(5, 5);
    pc.the_xyzrgba_pcl_pointcloud = pcl::PointCloud<pcl::PointXYZRGBA>(7, 7);

    pc2.the_xyzrgb_pcl_pointcloud = pcl::PointCloud<pcl::PointXYZRGB>(4, 4);
    pc2.the_xyzrgba_pcl_pointcloud = pcl::PointCloud<pcl::PointXYZRGBA>(3, 3);

    runTestWithInstances<PointCloudTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronEigenQuaternionTest)
{
    std::cout << "Running EigenQuaternion test" << std::endl;
    EigenQuaternionTest eq;
    EigenQuaternionTest eq2;
    runTestWithInstances<EigenQuaternionTest>(eq, eq2);
}

BOOST_AUTO_TEST_CASE(AronPositionTest)
{
    std::cout << "Running Position test" << std::endl;
    PositionTest pc;
    PositionTest pc2;
    runTestWithInstances<PositionTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronOrientationTest)
{
    std::cout << "Running Orientation test" << std::endl;
    OrientationTest pc;
    OrientationTest pc2;
    runTestWithInstances<OrientationTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronPoseTest)
{
    std::cout << "Running Pose test" << std::endl;
    PoseTest pc;
    PoseTest pc2;
    runTestWithInstances<PoseTest>(pc, pc2);
}

BOOST_AUTO_TEST_CASE(AronOptionalTest)
{
    std::cout << "Running Optional test" << std::endl;
    OptionalTest pc;
    OptionalTest pc2;
    runTestWithInstances<OptionalTest>(pc, pc2);
}
